import string
import tempfile
from io import BytesIO

import pandas as pd
import numpy as np

import pandas.testing as pt
import pytest
from mongoengine import Document, connect

from antarctic.PandasFields import SeriesField, FrameField, ParquetFrameField
from test.config import read_pd

#from mongomock.gridfs import enable_gridfs_integration
#enable_gridfs_integration()

client = connect(db="test", host="mongomock://localhost")


@pytest.fixture
def ts():
    return read_pd("ts.csv", squeeze=True, index_col=0, parse_dates=True)


@pytest.fixture
def prices():
    return read_pd("price.csv", index_col=0, parse_dates=True)


class Symbol(Document):
    close = SeriesField()
    prices = FrameField()
    #weights = FrameFileField()
    #ohlc = OhlcField()


def test_series(ts):
    s = Symbol()
    s.close = ts
    s.save()
    pt.assert_series_equal(s.close, ts)


def test_series_init(ts):
    s = Symbol(close=ts).save()
    pt.assert_series_equal(s.close, ts)


def test_not_a_series():
    s = Symbol()
    with pytest.raises(AssertionError):
        s.close = 6.0


def test_frame(prices):
    s = Symbol()
    s.prices = prices
    pt.assert_frame_equal(s.prices, prices)


def test_frame_init(prices):
    s = Symbol(prices=prices).save()
    pt.assert_frame_equal(s.prices, prices)


def test_not_a_frame():
    s = Symbol()
    with pytest.raises(AssertionError):
        s.prices = 2.0


def test_frame_series(ts, prices):
    s = Symbol(close=ts, prices=prices)

    pt.assert_frame_equal(s.prices, prices)
    pt.assert_series_equal(s.close, ts)


def test_save(ts, prices):
    s = Symbol(close=ts, prices=prices).save()
    pt.assert_frame_equal(s.prices, prices)
    pt.assert_series_equal(s.close, ts)


# def test_fileField(prices):
#     s = Symbol()
#     s.weights = prices
#     pt.assert_frame_equal(s.weights, prices)
#     s.save()


# def test_fileField_init(prices):
#     s = Symbol(weights=prices).save()
#     pt.assert_frame_equal(s.weights, prices)
#
#
# def test_not_a_FileFrame():
#     s = Symbol()
#     with pytest.raises(AssertionError):
#         s.weights = 2.0


# def test_ohlc_field():
#     s = Symbol()
#     ohlc = read_pd("ohlc.csv", index_col="time", parse_dates=True)
#     s.ohlc = ohlc
#     pt.assert_frame_equal(s.ohlc, ohlc)
#
#     print(s.ohlc.resample(rule="5min").last())
#     print(s.ohlc.resample(rule="5min").apply(lambda x: x.tail(1)))
#     print(s.ohlc.resample(rule="5min").apply(lambda x: x.head(1)))
#     # assert False
#
#     x = OhlcField.resample(frame=s.ohlc, rule="5min")
#     pt.assert_frame_equal(x, read_pd("ohlc_resample.csv", index_col="time", parse_dates=True))


def test_parquet_file():
    ohlc = read_pd("ohlc.csv", index_col="time", parse_dates=True)

    with tempfile.NamedTemporaryFile() as temp:
        ohlc.to_parquet(temp.name, engine='auto', compression=None)
        r = pd.read_parquet(temp.name, engine='auto')

        pt.assert_frame_equal(ohlc, r)


def test_parquet_bytes_io():
    # https://github.com/pandas-dev/pandas/issues/34467
    # This does not work with pandas 1.0.4!
    ohlc = read_pd("ohlc.csv", index_col="time", parse_dates=True)

    buffer = BytesIO()
    ohlc.to_parquet(buffer, engine='auto', compression=None)

    # don't write buffer to disc
    r = pd.read_parquet(buffer, engine='auto')
    pt.assert_frame_equal(ohlc, r)


def test_parquet_field():
    class Maffay(Document):
        frame = ParquetFrameField(engine="pyarrow", compression="gzip")

    maffay = Maffay()
    ohlc = read_pd("ohlc.csv", index_col="time", parse_dates=True)
    maffay.frame = ohlc

    pt.assert_frame_equal(maffay.frame, ohlc)


def test_parquet_field_large():
    class Maffay(Document):
        frame = ParquetFrameField(engine="pyarrow", compression=None)

    maffay = Maffay()

    # create random data
    def name():
        return "".join(np.random.choice(list(string.ascii_lowercase), size=10))

    ohlc = pd.DataFrame(data=np.random.randn(20000, 500), columns=[name() for i in range(0, 500)])

    maffay.frame = ohlc

    pt.assert_frame_equal(maffay.frame, ohlc)


def test_frame_field_large():
    class Maffay(Document):
        frame = FrameField()

    maffay = Maffay()

    # create random data
    def name():
        return "".join(np.random.choice(list(string.ascii_lowercase), size=10))

    ohlc = pd.DataFrame(data=np.random.randn(2000, 50), columns=[name() for i in range(0, 50)])

    maffay.frame = ohlc

    pt.assert_frame_equal(maffay.frame, ohlc)



import numpy as np
import pandas as pd
import pandas.testing as pt

import pytest
from mongoengine import connect, NotUniqueError
#from mongomock.gridfs import enable_gridfs_integration

from antarctic.Document import XDocument
from antarctic.PandasFields import SeriesField
from test.config import resource, read_pd

#enable_gridfs_integration()

client = connect(db="test", host="mongomock://localhost")


class Singer(XDocument):
    price = SeriesField()


def test_reference_frame():
    # can't harm to clean a bit
    Singer.objects.delete()

    p1 = Singer(name="Peter")
    p1.reference["A"] = 20.0
    p1.save()

    p2 = Singer(name="Falco")
    p2.reference["A"] = 30.0
    p2.reference["B"] = 10.0
    p2.save()

    frame = Singer.reference_frame(objects=[p1, p2])
    assert set(frame.index) == {"Peter", "Falco"}
    assert set(frame.keys()) == {"A", "B"}

    frame = Singer.reference_frame()
    assert set(frame.index) == {"Peter", "Falco"}
    assert set(frame.keys()) == {"A", "B"}


def test_lt():
    assert Singer(name="A") < Singer(name="B")


def test_reference():
    p = Singer(name="Peter Maffay")
    assert p.reference.get("NoNoNo", default=5) == 5

    p.reference["XXX"] = 10
    assert p.reference.keys() == {"XXX"}
    assert {k: v for k, v in p.reference.items()} == {"XXX": 10}


def test_equals():
    # can't harm to clean a bit
    Singer.objects.delete()

    p1 = Singer(name="Peter Maffay")
    p2 = Singer(name="Peter Maffay")

    assert p1 == p2


def test_products():
    # can't harm to clean a bit
    Singer.objects.delete()

    p1 = Singer(name="Peter").save()
    p2 = Singer(name="Falco").save()

    # here we query the database! Hence need the client in the background
    a = Singer.subset(names=["Peter"])
    assert len(a) == 1
    assert a[0] == p1

    b = Singer.subset()
    assert len(b) == 2
    assert set(b) == {p1, p2}

    frame = Singer.reference_frame(objects=[p1, p2])
    assert set(frame.index) == {"Peter", "Falco"}
    assert frame.empty

    frame = Singer.reference_frame()
    assert set(frame.index) == {"Peter", "Falco"}
    assert frame.empty


def test_repr():
    p1 = Singer(name="Peter Maffay")
    assert str(p1) == "<Singer: Peter Maffay>"


def test_not_unique_name():
    # can't harm to clean a bit
    Singer.objects.delete()

    # check that no singer has survived
    s = [singer for singer in Singer.objects]
    assert len(s) == 0

    c1 = Singer(name="AA").save()
    assert c1.name == "AA"

    # try to create a second singer with the same name
    with pytest.raises(NotUniqueError):
        Singer(name="AA").save()


def test_to_dict():
    # can't harm to clean a bit
    Singer.objects.delete()

    c1 = Singer(name="AAA").save()
    c2 = Singer(name="BBB").save()

    assert Singer.to_dict() == {"AAA": c1, "BBB": c2}


def test_apply():
    Singer.objects.delete()
    s1 = Singer(name="Falco").save()
    s2 = Singer(name="Peter Maffay").save()

    s1.price = pd.Series(index=[1, 2, 3], data=[7.0, 9.0, 8.0])
    s2.price = pd.Series(index=[1, 3], data=[8.0, 10.0])
    s3 = Singer(name="Karel Gott").save()
    s1.save()
    s2.save()

    a = pd.Series({name: value for name, value in Singer.apply(f=lambda x: x.price.mean(), default=np.nan)}).dropna()
    pt.assert_series_equal(a, pd.Series({"Falco": 8.0, "Peter Maffay": 9.0}))


def test_repr():
    Singer.objects.delete()
    s1 = Singer(name="Falco").save()
    assert str(s1) == '<Singer: Falco>'
    assert s1.__repr__() == '<Singer: Falco>'


def test_frame():
    Singer.objects.delete()
    s1 = Singer(name="Falco").save()
    s2 = Singer(name="Peter Maffay").save()
    s3 = Singer(name="Karel Gott").save()

    s1.price = pd.Series(index=[1, 2, 3], data=[7.0, 9.0, 8.0])
    s2.price = pd.Series(index=[1, 3], data=[8.0, 10.0])
    s1.save()
    s2.save()

    f = Singer.frame(series="price")
    pt.assert_frame_equal(f, read_pd("frame.csv", index_col=0))

    with pytest.raises(AttributeError):
        Singer.frame(series="wurst")


def test_names():
    Singer.objects.delete()
    s1 = Singer(name="A").save()
    s2 = Singer(name="B").save()
    s3 = Singer(name="C").save()

    assert {s1, s2} == set(Singer.objects(name__in=["A", "B"]).all())
    assert {s2, s1} == set(Singer.objects(name__in=["B", "A"]).all())


from io import BytesIO

import pandas as pd
from mongoengine.base import BaseField


class SeriesField(BaseField):
    def __set__(self, instance, value):
        # convert the incoming series into a json document
        if value is not None:
            # check it's really a series
            if not isinstance(value, str):
                assert isinstance(value, pd.Series)
                # convert the series into a json string
                value = value.to_json(orient="split")

        # give the (new) value to mum
        super(SeriesField, self).__set__(instance, value)

    def __get__(self, instance, owner):
        # ask mum for the value stored
        x = super(SeriesField, self).__get__(instance, owner)

        if x is not None:
            # convert the value into a series
            return pd.read_json(x, orient="split", typ="series")

        return None


class FrameField(BaseField):
    def __set__(self, instance, value):
        # convert the incoming series into a json document
        if value is not None:
            # check it's really a DataFrame
            if not isinstance(value, str):
                assert isinstance(value, pd.DataFrame)
                # convert the frame into a json string
                value = value.to_json(orient="table")

        # give the (new) value to mum
        super(FrameField, self).__set__(instance, value)

    def __get__(self, instance, owner):
        # ask mum for the value stored
        x = super(FrameField, self).__get__(instance, owner)

        if x is not None:
            return pd.read_json(x, orient="table", typ="frame")

        return None


class ParquetFrameField(BaseField):
    def __init__(self, engine="auto", compression=None, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine
        self.compression = compression

    def __set__(self, instance, value):
        # convert the incoming series into a json document
        if value is not None:
            # check it's really a DataFrame
            if not isinstance(value, str):
                assert isinstance(value, pd.DataFrame)
                # convert the frame into a json string
                with BytesIO() as buffer:
                    value.to_parquet(buffer, engine=self.engine, compression=self.compression)
                    value = buffer.getvalue()

        # give the (new) value to mum
        super(ParquetFrameField, self).__set__(instance, value)

    def __get__(self, instance, owner):
        # ask mum for the value stored
        x = super(ParquetFrameField, self).__get__(instance, owner)

        if x is not None:
            with BytesIO(x) as buffer:
                return pd.read_parquet(buffer, engine=self.engine)

        return None


# class OhlcField(FrameField):
#     def __set__(self, instance, value):
#         if isinstance(value, pd.DataFrame):
#             assert {"open", "high", "low", "close", "volume"}.issubset(set(value.keys()))
#
#         super(OhlcField, self).__set__(instance, value)
#
#     def __get__(self, instance, owner):
#         x = super(OhlcField, self).__get__(instance, owner)
#
#         if x is not None:
#             assert {"open", "high", "low", "close", "volume"}.issubset(set(x.keys()))
#
#         return x
#
#     @classmethod
#     def resample(cls, frame, rule):
#         d = {"open": lambda x: x.head(1),
#              "high": lambda x: x.max(),
#              "low": lambda x: x.min(),
#              "close": lambda x: x.tail(1),
#              "volume": lambda x: x.sum()}
#
#         return pd.DataFrame({name: frame[name].resample(rule, loffset=rule).apply(f) for name, f in d.items()})


# class FrameFileField(FileField):
#     def __set__(self, instance, value):
#         # convert the incoming series into a json document
#         if value is not None:
#             # check it's really a DataFrame
#             assert isinstance(value, pd.DataFrame)
#             # convert the frame into a json string
#             value = value.to_json(orient="table").encode()
#
#         # give the (new) value to mum
#         super(FrameFileField, self).__set__(instance, value)
#
#     def __get__(self, instance, owner):
#         # ask mum for the value stored
#         x = super(FrameFileField, self).__get__(instance, owner).read().decode()
#         return pd.read_json(x, typ="frame", orient="table")



from datetime import datetime

import pandas as pd
from mongoengine import *


class XDocument(Document):
    # A XDocument is an abstract Mongo Document, e.g. instances of this document can not be instantiated.
    # All concrete objects such as Symbols or Strategies are children of the XDocument.
    # Having a common parent helps to share functionality
    meta = {'abstract': True}

    name = StringField(unique=True, required=True)
    reference = DictField()

    # Date modified
    date_modified = DateTimeField(default=datetime.utcnow)

    @classmethod
    def reference_frame(cls, objects=None) -> pd.DataFrame:
        objects = objects or cls.objects

        frame = pd.DataFrame(
            {obj.name: pd.Series({key: data for key, data in obj.reference.items()}, dtype=object) for obj in
             objects}).transpose()
        frame.index.name = cls.__name__.lower()
        return frame.sort_index()

    @classmethod
    def subset(cls, names=None):
        # extract symbols from database
        if names is None:
            return cls.objects
        else:
            return cls.objects(name__in=names)

    @classmethod
    def to_dict(cls, objects=None):
        # represent all documents of a class as a dictionary
        objects = objects or cls.objects
        return {x.name: x for x in objects}

    @classmethod
    def apply(cls, f, default, objects=None) -> pd.DataFrame:
        objects = objects or cls.objects

        for obj in objects:
            try:
                yield obj.name, f(obj)
            except (AttributeError, KeyError):
                yield obj.name, default

    @classmethod
    def frame(cls, series, objects=None) -> pd.DataFrame:
        objects = objects or cls.objects
        return pd.DataFrame({p.name: p.__getattribute__(series) for p in objects}).dropna(axis=1, how="all")

    def __lt__(self, other):
        # sort documents by name
        return self.name < other.name

    def __eq__(self, other):
        # two documents are the sname if they have the same name and class
        return self.__class__ == other.__class__ and self.name == other.name

    # we want to make a set of assets, etc....
    def __hash__(self):
        return hash(self.to_json())

    def __str__(self):
        return "<{type}: {name}>".format(type=self.__class__.__name__, name=self.name)

    def __repr__(self):
        return "<{type}: {name}>".format(type=self.__class__.__name__, name=self.name)